@@ -2,6 +2,7 @@ package graphql_test
 
 import (
 	"context"
+	"encoding/json"
 	"errors"
 	"fmt"
 	"sync"
@@ -10,6 +11,7 @@ import (
 
 	"github.com/graph-gophers/graphql-go"
 	gqlerrors "github.com/graph-gophers/graphql-go/errors"
+	"github.com/graph-gophers/graphql-go/example/social"
 	"github.com/graph-gophers/graphql-go/example/starwars"
 	"github.com/graph-gophers/graphql-go/gqltesting"
 	"github.com/graph-gophers/graphql-go/introspection"
@@ -48,6 +50,10 @@ func (r *helloSnakeResolver2) SayHello(ctx context.Context, args struct{ FullNam
 	return "Hello " + args.FullName + "!", nil
 }
 
+type structFieldResolver struct {
+	Hello string
+}
+
 type theNumberResolver struct {
 	number int32
 }
@@ -96,21 +102,12 @@ func (e resolverNotFoundError) Extensions() map[string]interface{} {
 	}
 }
 
-type findDroidResolver struct{}
-
-func (r *findDroidResolver) FindDroid(ctx context.Context) (string, error) {
-	return "", resolverNotFoundError{
-		Code:    "NotFound",
-		Message: "This is not the droid you are looking for",
-	}
-}
-
 var (
 	droidNotFoundError = resolverNotFoundError{
 		Code:    "NotFound",
 		Message: "This is not the droid you are looking for",
 	}
-	quoteError = errors.New("Bleep bloop")
+	errQuote = errors.New("bleep bloop")
 
 	r2d2          = &droidResolver{name: "R2-D2"}
 	c3po          = &droidResolver{name: "C-3PO"}
@@ -153,7 +150,7 @@ func (d *droidResolver) Name() (string, error) {
 func (d *droidResolver) Quotes() ([]string, error) {
 	switch d.name {
 	case r2d2.name:
-		return nil, quoteError
+		return nil, errQuote
 	case c3po.name:
 		return []string{"We're doomed!", "R2-D2, where are you?"}, nil
 	}
@@ -191,7 +188,6 @@ func TestHelloWorld(t *testing.T) {
 				}
 			`,
 		},
-
 		{
 			Schema: graphql.MustParseSchema(`
 				schema {
@@ -216,6 +212,36 @@ func TestHelloWorld(t *testing.T) {
 	})
 }
 
+func TestHelloWorldStructFieldResolver(t *testing.T) {
+	t.Parallel()
+
+	gqltesting.RunTests(t, []*gqltesting.Test{
+		{
+			Schema: graphql.MustParseSchema(`
+				schema {
+					query: Query
+				}
+
+				type Query {
+					hello: String!
+				}
+			`,
+				&structFieldResolver{Hello: "Hello world!"},
+				graphql.UseFieldResolvers()),
+			Query: `
+				{
+					hello
+				}
+			`,
+			ExpectedResult: `
+				{
+					"hello": "Hello world!"
+				}
+			`,
+		},
+	})
+}
+
 func TestHelloSnake(t *testing.T) {
 	t.Parallel()
 
@@ -586,7 +612,7 @@ func TestBasic(t *testing.T) {
 
 type testEmbeddedStructResolver struct{}
 
-func (_ *testEmbeddedStructResolver) Course() courseResolver {
+func (*testEmbeddedStructResolver) Course() courseResolver {
 	return courseResolver{
 		CourseMeta: CourseMeta{
 			Name:       "Biology",
@@ -907,8 +933,8 @@ func TestErrorPropagationInLists(t *testing.T) {
 			`,
 			ExpectedErrors: []*gqlerrors.QueryError{
 				{
-					Message:       quoteError.Error(),
-					ResolverError: quoteError,
+					Message:       errQuote.Error(),
+					ResolverError: errQuote,
 					Path:          []interface{}{"findDroids", 0, "quotes"},
 				},
 			},
@@ -942,8 +968,8 @@ func TestErrorPropagationInLists(t *testing.T) {
 			`,
 			ExpectedErrors: []*gqlerrors.QueryError{
 				{
-					Message:       quoteError.Error(),
-					ResolverError: quoteError,
+					Message:       errQuote.Error(),
+					ResolverError: errQuote,
 					Path:          []interface{}{"findNilDroids", 0, "quotes"},
 				},
 				{
@@ -1179,6 +1205,37 @@ func TestFragments(t *testing.T) {
 				}
 			`,
 		},
+		{
+			Schema: starwarsSchema,
+			Query: `
+				query {
+					human(id: "1000") {
+						id
+						mass
+						...characterInfo
+					}
+				}
+				fragment characterInfo on Character {
+					name
+					...on Droid {
+						primaryFunction
+					}
+					...on Human {
+						height
+					}
+				}
+			`,
+			ExpectedResult: `
+				{
+					"human": {
+						"id": "1000",
+						"mass": 77,
+						"name": "Luke Skywalker",
+						"height": 1.72
+					}
+				}
+			`,
+		},
 	})
 }
 
@@ -1824,7 +1881,7 @@ func TestInlineFragments(t *testing.T) {
 		},
 
 		{
-			Schema: socialSchema,
+			Schema: graphql.MustParseSchema(social.Schema, &social.Resolver{}, graphql.UseFieldResolvers()),
 			Query: `
 				query {
 					admin(id: "0x01") {
@@ -2213,7 +2270,6 @@ func TestIntrospection(t *testing.T) {
 							{ "name": "SearchResult" },
 							{ "name": "Starship" },
 							{ "name": "String" },
-							{ "name": "_Service" },
 							{ "name": "__Directive" },
 							{ "name": "__DirectiveLocation" },
 							{ "name": "__EnumValue" },
@@ -2859,7 +2915,7 @@ func TestTime(t *testing.T) {
 
 type resolverWithUnexportedMethod struct{}
 
-func (r *resolverWithUnexportedMethod) changeTheNumber(args struct{ NewNumber int32 }) int32 {
+func (r *resolverWithUnexportedMethod) changeTheNumber(args struct{ NewNumber int32 }) int32 { //lint:ignore U1000 ingore this for now
 	return args.NewNumber
 }
 
@@ -3027,7 +3083,7 @@ type recursive struct {
 }
 
 func (r *inputResolver) Recursive(args struct{ Value *recursive }) int32 {
-	n := int32(0)
+	var n int32
 	v := args.Value
 	for v != nil {
 		v = v.Next
@@ -3064,7 +3120,7 @@ func TestInput(t *testing.T) {
 			nullableIntEnumValue(value: IntEnum): IntEnum
 			intEnum(value: IntEnum!): IntEnum!
 			nullableIntEnum(value: IntEnum): IntEnum
-			recursive(value: RecursiveInput!): Int!
+			recursive(value: RecursiveInput): Int!
 			id(value: ID!): ID!
 		}
 
@@ -3409,7 +3465,7 @@ func TestComposedFragments(t *testing.T) {
 }
 
 var (
-	exampleError = fmt.Errorf("This is an error")
+	errExample = fmt.Errorf("this is an error")
 
 	nilChildErrorString = `graphql: got nil for non-null "Child"`
 )
@@ -3417,7 +3473,7 @@ var (
 type childResolver struct{}
 
 func (r *childResolver) TriggerError() (string, error) {
-	return "This will never be returned to the client", exampleError
+	return "This will never be returned to the client", errExample
 }
 func (r *childResolver) NoError() string {
 	return "no error"
@@ -3455,8 +3511,8 @@ func TestErrorPropagation(t *testing.T) {
 			`,
 			ExpectedErrors: []*gqlerrors.QueryError{
 				{
-					Message:       exampleError.Error(),
-					ResolverError: exampleError,
+					Message:       errExample.Error(),
+					ResolverError: errExample,
 					Path:          []interface{}{"triggerError"},
 				},
 			},
@@ -3494,8 +3550,8 @@ func TestErrorPropagation(t *testing.T) {
 			`,
 			ExpectedErrors: []*gqlerrors.QueryError{
 				{
-					Message:       exampleError.Error(),
-					ResolverError: exampleError,
+					Message:       errExample.Error(),
+					ResolverError: errExample,
 					Path:          []interface{}{"child", "triggerError"},
 				},
 			},
@@ -3537,8 +3593,8 @@ func TestErrorPropagation(t *testing.T) {
 			`,
 			ExpectedErrors: []*gqlerrors.QueryError{
 				{
-					Message:       exampleError.Error(),
-					ResolverError: exampleError,
+					Message:       errExample.Error(),
+					ResolverError: errExample,
 					Path:          []interface{}{"child", "child", "triggerError"},
 				},
 			},
@@ -3583,8 +3639,8 @@ func TestErrorPropagation(t *testing.T) {
 			`,
 			ExpectedErrors: []*gqlerrors.QueryError{
 				{
-					Message:       exampleError.Error(),
-					ResolverError: exampleError,
+					Message:       errExample.Error(),
+					ResolverError: errExample,
 					Path:          []interface{}{"child", "child", "triggerError"},
 				},
 			},
@@ -3708,8 +3764,8 @@ func TestErrorPropagation(t *testing.T) {
 					Path:    []interface{}{"child", "child", "child", "nilChild"},
 				},
 				{
-					Message:       exampleError.Error(),
-					ResolverError: exampleError,
+					Message:       errExample.Error(),
+					ResolverError: errExample,
 					Path:          []interface{}{"child", "child", "triggerError"},
 				},
 			},
@@ -4507,8 +4563,8 @@ func TestCircularFragmentMaxDepth(t *testing.T) {
 	              }
 	          `,
 			ExpectedErrors: []*gqlerrors.QueryError{{
-				Message: `Cannot spread fragment "X" within itself via Y.`,
-				Rule:    "NoFragmentCycles",
+				Message: `Cannot spread fragment "X" within itself via "Y".`,
+				Rule:    "NoFragmentCyclesRule",
 				Locations: []gqlerrors.Location{
 					{Line: 7, Column: 20},
 					{Line: 10, Column: 20},
@@ -4518,33 +4574,504 @@ func TestCircularFragmentMaxDepth(t *testing.T) {
 	})
 }
 
-func TestQueryService(t *testing.T) {
+func TestMaxQueryLength(t *testing.T) {
+	withMaxQueryLen := graphql.MustParseSchema(starwars.Schema, &starwars.Resolver{}, graphql.MaxQueryLength(75))
+	gqltesting.RunTests(t, []*gqltesting.Test{
+		{
+			Schema: withMaxQueryLen,
+			// Query length is 69 bytes
+			Query: `
+				query {
+					hero(episode: EMPIRE) {
+						name
+					}
+				}
+			`,
+			ExpectedResult: `{"hero":{"name":"Luke Skywalker"}}`,
+		},
+		{
+			Schema: withMaxQueryLen,
+			Query: `
+				query HeroForEpisode {
+					hero(episode: WRATH_OF_KHAN) {
+						name
+					}
+				}
+			`,
+			ExpectedErrors: []*gqlerrors.QueryError{{
+				Message: `query length 91 exceeds the maximum allowed query length of 75 bytes`,
+			}},
+		},
+	})
+}
+
+type RootResolver struct{}
+type QueryResolver struct{}
+type MutationResolver struct{}
+type SubscriptionResolver struct {
+	err      error
+	upstream <-chan *helloEventResolver
+}
+
+func (r *RootResolver) Query() *QueryResolver {
+	return &QueryResolver{}
+}
+
+func (r *RootResolver) Mutation() *MutationResolver {
+	return &MutationResolver{}
+}
+
+type helloEventResolver struct {
+	msg string
+	err error
+}
+
+func (r *helloEventResolver) Msg() (string, error) {
+	return r.msg, r.err
+}
+
+func closedHelloEventUpstream(rr ...*helloEventResolver) <-chan *helloEventResolver {
+	c := make(chan *helloEventResolver, len(rr))
+	for _, r := range rr {
+		c <- r
+	}
+	close(c)
+	return c
+}
+
+func (r *RootResolver) Subscription() *SubscriptionResolver {
+	return &SubscriptionResolver{
+		upstream: closedHelloEventUpstream(
+			&helloEventResolver{msg: "Hello subscription!"},
+			&helloEventResolver{err: errors.New("resolver error")},
+			&helloEventResolver{msg: "Hello again!"},
+		),
+	}
+}
+
+func (qr *QueryResolver) Hello() string {
+	return "Hello query!"
+}
+
+func (mr *MutationResolver) Hello() string {
+	return "Hello mutation!"
+}
+
+func (sr *SubscriptionResolver) Hello(ctx context.Context) (chan *helloEventResolver, error) {
+	if sr.err != nil {
+		return nil, sr.err
+	}
+
+	c := make(chan *helloEventResolver)
+	go func() {
+		for r := range sr.upstream {
+			select {
+			case <-ctx.Done():
+				close(c)
+				return
+			case c <- r:
+			}
+		}
+		close(c)
+	}()
+
+	return c, nil
+}
+
+type errRootResolver1 struct {
+	RootResolver
+}
+
+// Query is invalid because it doesn't have a return value.
+func (*errRootResolver1) Query() {}
+
+type errRootResolver2 struct {
+	RootResolver
+}
+
+// Query is invalid because it has more than 1 return value
+func (*errRootResolver2) Query() (*QueryResolver, error) {
+	return nil, nil
+}
+
+type errRootResolver3 struct {
+	RootResolver
+}
+
+// Mutation is invalid because it returns nil
+func (*errRootResolver3) Mutation() *MutationResolver {
+	return nil
+}
+
+type errRootResolver4 struct {
+	RootResolver
+}
+
+// Query is invalid because it doesn't return a pointer.
+func (*errRootResolver4) Query() MutationResolver {
+	return MutationResolver{}
+}
+
+type errRootResolver5 struct {
+	RootResolver
+}
+
+// Query is invalid because it returns *[]int instead of a resolver.
+func (*errRootResolver5) Query() *[]int {
+	return &[]int{1, 2}
+}
+
+type errRootResolver6 struct {
+	RootResolver
+}
+
+// Mutation is invalid because it returns a map[string]int instead of a resolver.
+func (*errRootResolver6) Mutation() map[string]int {
+	return map[string]int{"key": 3}
+}
+
+type errRootResolver7 struct {
+	RootResolver
+}
+
+// Subscription is invalid because it returns an invalid resolver.
+func (*errRootResolver7) Subscription() interface{} {
+	a := struct {
+		Name string
+	}{Name: "invalid"}
+	return &a
+}
+
+type errRootResolver8 struct {
+	RootResolver
+}
+
+// Query is invalid because it accepts arguments.
+func (*errRootResolver8) Query(ctx context.Context) *QueryResolver {
+	return &QueryResolver{}
+}
+
+// TestSeparateResolvers ensures that a field with the same name is allowed in different operations
+func TestSeparateResolvers(t *testing.T) {
+	helloEverywhere := `
+		schema {
+			query: Query
+			mutation: Mutation
+			subscription: Subscription
+		}
+
+		type Query {
+			hello: String!
+		}
+
+		type Mutation {
+			hello: String!
+		}
+
+		type Subscription {
+			hello: HelloEvent!
+		}
+
+		type HelloEvent {
+			msg: String!
+		}
+	`
+
+	separateSchema := graphql.MustParseSchema(helloEverywhere, &RootResolver{})
+
+	gqltesting.RunTests(t, []*gqltesting.Test{
+		{
+			Schema: separateSchema,
+			Query: `
+				query {
+					hello
+				}
+			`,
+			ExpectedResult: `
+				{
+					"hello": "Hello query!"
+				}
+			`,
+		},
+		{
+			Schema: separateSchema,
+			Query: `
+				mutation {
+					hello
+				}
+			`,
+			ExpectedResult: `
+				{
+					"hello": "Hello mutation!"
+				}
+			`,
+		},
+	})
+
+	gqltesting.RunSubscribes(t, []*gqltesting.TestSubscription{
+		{
+			Name:   "ok",
+			Schema: separateSchema,
+			Query: `
+				subscription {
+					hello {
+						msg
+					}
+				}
+			`,
+			ExpectedResults: []gqltesting.TestResponse{
+				{
+					Data: json.RawMessage(`
+						{
+							"hello": {
+								"msg": "Hello subscription!"
+							}
+						}
+					`),
+				},
+				{
+					// null propagates all the way up because msg is non-null
+					Data:   json.RawMessage(`null`),
+					Errors: []*gqlerrors.QueryError{gqlerrors.Errorf("%s", errResolver)},
+				},
+				{
+					Data: json.RawMessage(`
+						{
+							"hello": {
+								"msg": "Hello again!"
+							}
+						}
+					`),
+				},
+			},
+		},
+	})
+
+	// test errors with invalid resolvers
+	tests := []struct {
+		name     string
+		resolver interface{}
+		opts     []graphql.SchemaOpt
+		wantErr  string
+	}{
+		{
+			name:     "query_method_has_no_return_val",
+			resolver: &errRootResolver1{},
+			wantErr:  "method \"Query\" of *graphql_test.errRootResolver1 must have 1 return value, got 0",
+		},
+		{
+			name:     "query_method_returns_too_many_vals",
+			resolver: &errRootResolver2{},
+			wantErr:  "method \"Query\" of *graphql_test.errRootResolver2 must have 1 return value, got 2",
+		},
+		{
+			name:     "mutation_method_returns_nil",
+			resolver: &errRootResolver3{},
+			wantErr:  "method \"Mutation\" of *graphql_test.errRootResolver3 must return a non-nil result, got <nil>",
+		},
+		{
+			name:     "query_method_does_not_return_a_pointer",
+			resolver: &errRootResolver4{},
+			wantErr:  "method \"Query\" of *graphql_test.errRootResolver4 must return an interface or a pointer, got graphql_test.MutationResolver",
+		},
+		{
+			name:     "query_method_returns_invalid_resolver_type",
+			resolver: &errRootResolver5{},
+			wantErr:  "*[]int does not resolve \"Query\": missing method for field \"hello\"",
+		},
+		{
+			name:     "mutation_method_returns_invalid_resolver_type",
+			resolver: &errRootResolver6{},
+			wantErr:  "method \"Mutation\" of *graphql_test.errRootResolver6 must return an interface or a pointer, got map[string]int",
+		},
+		{
+			name:     "query_subscription_returns_invalid_resolver_type",
+			resolver: &errRootResolver7{},
+			wantErr:  "*struct { Name string } does not resolve \"Subscription\": missing method for field \"hello\"",
+		},
+		{
+			name:     "mutation_method_returns_invalid_resolver_type",
+			resolver: &errRootResolver8{},
+			wantErr:  "method \"Query\" of *graphql_test.errRootResolver8 must not accept any arguments, got 1",
+		},
+	}
+	for _, tt := range tests {
+		t.Run(tt.name, func(t *testing.T) {
+			_, err := graphql.ParseSchema(helloEverywhere, tt.resolver, tt.opts...)
+			if err == nil {
+				t.Fatalf("want err: %q, got: <nil>", tt.wantErr)
+			}
+			if err.Error() != tt.wantErr {
+				t.Fatalf("want err: %q, got: %q", tt.wantErr, err.Error())
+			}
+		})
+	}
+}
+
+func TestSchemaExtension(t *testing.T) {
 	t.Parallel()
 
-	schemaString := `
+	sdl := `
+	directive @awesome on SCHEMA
+
 	schema {
 		query: Query
 	}
 
 	type Query {
 		hello: String!
-	}`
+	}
+	
+	extend schema @awesome
+	`
+	schema := graphql.MustParseSchema(sdl, &helloResolver{})
 
 	gqltesting.RunTests(t, []*gqltesting.Test{
 		{
-			Schema: graphql.MustParseSchema(schemaString, &helloWorldResolver1{}),
+			Schema: schema,
 			Query: `
 				{
-					_service{
-						sdl
-					}
+					hello
 				}
 			`,
 			ExpectedResult: `
 				{
-					"_service": {
-						"sdl": "\n\tschema {\n\t\tquery: Query\n\t}\n\n\ttype Query {\n\t\thello: String!\n\t}"
+					"hello": "Hello world!"
+				}
+			`,
+		},
+	})
+
+	ast := schema.AST()
+	dirs := ast.SchemaDefinition.Directives
+	if len(dirs) != 1 {
+		t.Fatalf("expected 1 schema directive, got %d", len(dirs))
+	}
+	name := dirs[0].Name.Name
+	if name != "awesome" {
+		t.Fatalf(`expected an "awesome" schema directive, got %q`, dirs[0].Name.Name)
+	}
+}
+
+func TestGraphqlNames(t *testing.T) {
+	t.Parallel()
+
+	sdl1 := `
+	type Query {
+		hello: String!
+	}
+	`
+	type invalidResolver1 struct {
+		Field1 string `graphql:"hello"`
+		Field2 string `graphql:"hello"`
+	}
+
+	wantErr := fmt.Errorf(`*graphql_test.invalidResolver1 does not resolve "Query": multiple fields have a graphql reflect tag "hello"`)
+	_, err := graphql.ParseSchema(sdl1, &invalidResolver1{}, graphql.UseFieldResolvers())
+	if err == nil || err.Error() != wantErr.Error() {
+		t.Fatalf("want err %q, got %q", wantErr, err)
+	}
+
+	gqltesting.RunTests(t, []*gqltesting.Test{
+		{
+			Schema: graphql.MustParseSchema(`
+				type Query {
+					_hello: String!
+					hello: String!
+					Hello: String!
+					HELLO: String!
+				}`,
+				func() interface{} {
+					type helloTagResolver struct {
+						Hello           string
+						HelloUnderscore string `graphql:"_hello"`
+						HelloLower      string `graphql:"hello"`
+						HelloTitle      string `graphql:"Hello"`
+						HelloUpper      string `graphql:"HELLO"`
 					}
+					return &helloTagResolver{
+						Hello:           "This field will not be used during query execution!",
+						HelloLower:      "Hello, graphql!",
+						HelloTitle:      "Hello, GraphQL!",
+						HelloUnderscore: "Hello, _!",
+						HelloUpper:      "Hello, GRAPHQL!",
+					}
+				}(),
+				graphql.UseFieldResolvers()),
+			Query: `
+				{
+					_hello
+					hello
+					Hello
+					HELLO
+				}
+			`,
+			ExpectedResult: `
+				{
+					"_hello": "Hello, _!",
+				    "hello": "Hello, graphql!",
+				    "Hello": "Hello, GraphQL!",
+				    "HELLO": "Hello, GRAPHQL!"
+				}
+			`,
+		},
+	})
+}
+
+func Test_fieldFunc(t *testing.T) {
+	sdl := `
+		type Query {
+			hello(name: String!): String!
+		}
+	`
+	gqltesting.RunTests(t, []*gqltesting.Test{
+		{
+			Schema: graphql.MustParseSchema(sdl,
+				func() interface{} {
+					type helloTagResolver struct {
+						Hello func(args struct{ Name string }) string
+					}
+					fn := func(args struct{ Name string }) string {
+						return "Hello, " + args.Name + "!"
+					}
+					return &helloTagResolver{
+						Hello: fn,
+					}
+				}(),
+				graphql.UseFieldResolvers()),
+			Query: `
+				{
+					hello(name: "GraphQL")
+				}
+			`,
+			ExpectedResult: `
+				{
+				    "hello": "Hello, GraphQL!"
+				}
+			`,
+		},
+		{
+			Schema: graphql.MustParseSchema(sdl,
+				func() interface{} {
+					type helloTagResolver struct {
+						Greet func(ctx context.Context, args struct{ Name string }) (string, error) `graphql:"hello"`
+					}
+					fn := func(_ context.Context, args struct{ Name string }) (string, error) {
+						return "Hello, " + args.Name + "!", nil
+					}
+					return &helloTagResolver{
+						Greet: fn,
+					}
+				}(),
+				graphql.UseFieldResolvers()),
+			Query: `
+				{
+					hello(name: "GraphQL")
+				}
+			`,
+			ExpectedResult: `
+				{
+				    "hello": "Hello, GraphQL!"
 				}
 			`,
 		},
