@@ -3,8 +3,8 @@ package resolvable
 import (
 	"reflect"
 
+	"github.com/graph-gophers/graphql-go/ast"
 	"github.com/graph-gophers/graphql-go/introspection"
-	"github.com/graph-gophers/graphql-go/types"
 )
 
 // Meta defines the details of the metadata schema for introspection.
@@ -18,24 +18,18 @@ type Meta struct {
 	Service       *Object
 }
 
-func newMeta(s *types.Schema) *Meta {
+func newMeta(s *ast.Schema) *Meta {
 	var err error
-	b := newBuilder(s)
+	b := newBuilder(s, false)
 
-	metaSchema := s.Types["__Schema"].(*types.ObjectTypeDefinition)
-	so, err := b.makeObjectExec(metaSchema.Name, metaSchema.Fields, nil, false, reflect.TypeOf(&introspection.Schema{}))
+	metaSchema := s.Types["__Schema"].(*ast.ObjectTypeDefinition)
+	so, err := b.makeObjectExec(metaSchema.Name, metaSchema.Fields, nil, nil, false, reflect.TypeOf(&introspection.Schema{}))
 	if err != nil {
 		panic(err)
 	}
 
-	metaType := s.Types["__Type"].(*types.ObjectTypeDefinition)
-	t, err := b.makeObjectExec(metaType.Name, metaType.Fields, nil, false, reflect.TypeOf(&introspection.Type{}))
-	if err != nil {
-		panic(err)
-	}
-
-	metaService := s.Types["_Service"].(*types.ObjectTypeDefinition)
-	sv, err := b.makeObjectExec(metaService.Name, metaService.Fields, nil, false, reflect.TypeOf(&introspection.Service{}))
+	metaType := s.Types["__Type"].(*ast.ObjectTypeDefinition)
+	t, err := b.makeObjectExec(metaType.Name, metaType.Fields, nil, nil, false, reflect.TypeOf(&introspection.Type{}))
 	if err != nil {
 		panic(err)
 	}
@@ -45,15 +39,15 @@ func newMeta(s *types.Schema) *Meta {
 	}
 
 	fieldTypename := Field{
-		FieldDefinition: types.FieldDefinition{
+		FieldDefinition: ast.FieldDefinition{
 			Name: "__typename",
-			Type: &types.NonNull{OfType: s.Types["String"]},
+			Type: &ast.NonNull{OfType: s.Types["String"]},
 		},
 		TraceLabel: "GraphQL field: __typename",
 	}
 
 	fieldSchema := Field{
-		FieldDefinition: types.FieldDefinition{
+		FieldDefinition: ast.FieldDefinition{
 			Name: "__schema",
 			Type: s.Types["__Schema"],
 		},
@@ -61,28 +55,18 @@ func newMeta(s *types.Schema) *Meta {
 	}
 
 	fieldType := Field{
-		FieldDefinition: types.FieldDefinition{
+		FieldDefinition: ast.FieldDefinition{
 			Name: "__type",
 			Type: s.Types["__Type"],
 		},
 		TraceLabel: "GraphQL field: __type",
 	}
 
-	fieldService := Field{
-		FieldDefinition: types.FieldDefinition{
-			Name: "_service",
-			Type: s.Types["_Service"],
-		},
-		TraceLabel: "GraphQL field: _service",
-	}
-
 	return &Meta{
 		FieldSchema:   fieldSchema,
 		FieldTypename: fieldTypename,
 		FieldType:     fieldType,
-		FieldService:  fieldService,
 		Schema:        so,
 		Type:          t,
-		Service:       sv,
 	}
 }
